from importlib import resources

from . import templates


class DownloadType:
    """Helper to select the download type"""

    ISSUED = "RfcEmisor"
    RECEIVED = "RfcReceptor"


class RequestType:
    """Helper to select the request type"""

    CFDI = "CFDI"
    METADATA = "Metadata"


TEMPLATES = {
    "Envelope": resources.read_text(templates.common, "Envelope.xml"),
    "KeyInfo": resources.read_text(templates.common, "KeyInfo.xml"),
    "Signature": resources.read_text(templates.common, "Signature.xml"),
    "SignedInfo": resources.read_text(templates.common, "SignedInfo.xml"),
    "Timestamp": resources.read_text(templates.login, "Timestamp.xml"),
    "LoginEnvelope": resources.read_text(templates.login, "Envelope.xml"),
    "SolicitaDescarga": resources.read_text(templates.query, "SolicitaDescarga.xml"),
    "VerificaSolicitudDescarga": resources.read_text(
        templates.verify, "VerificaSolicitudDescarga.xml"
    ),
    "PeticionDescargaMasivaTercerosEntrada": resources.read_text(
        templates.download, "PeticionDescargaMasivaTercerosEntrada.xml"
    ),
}



import logging
from datetime import datetime, timedelta
from uuid import uuid1

from . import exceptions, utils
from .certificate_handler import CertificateHandler
from .enums import TEMPLATES
from .sat_parsers import LoginParser

_logger = logging.getLogger(__name__)


class SATLoginHandler:
    _token: str = None
    _token_expires: datetime
    _certificate_handler: CertificateHandler

    def __init__(self, certificate_handler):
        self._certificate_handler = certificate_handler

    @property
    def token(self) -> str:
        self.login()
        return self._token

    def _token_expired(self) -> bool:
        """Checks if the token expiration date is yet to come

        Returns:
            bool: True if not token or if is expired
        """
        if not self._token or not self._token_expires or datetime.utcnow() > self._token_expires:
            _logger.debug("Token expired")
            return True
        return False

    def login(self, created: datetime = None, expires: datetime = None, uuid: str = None):
        """If the current token is invalid, tries to login

        Args:
            created (datetime, optional): Creation date to be used in the session.
                                          Defaults to datetime.utcnow().
            expires (datetime, optional): Expiration date to be used in the session.
                                          Defaults to datetime.utcnow()+timedelta(minutes=5).
            uuid (str, optional): UUID date to be used in the session.
                                  Defaults to f'uuid-{uuid1()}-1'.
        """
        if not self._token_expired():
            return
        _logger.debug("Token expired, creating a new one")
        created = created or datetime.utcnow()
        expires = expires or created + timedelta(minutes=5)
        uuid = uuid or f"uuid-{uuid1()}-1"
        self._token_expires = expires
        self._login(created, expires, uuid)
        _logger.debug("New token created")

    def _login(self, created: datetime, expires: datetime, uuid: str):
        """Send login request to the SAT

        Args:
            created (datetime): Creation date to be used in the session
            expires (datetime): Expiration date to be used in the session
            uuid (str): UUID date to be used in the session

        Raises:
            RequestException: If there was an error in the request
        """
        request_content = self._get_login_soap_body(created, expires, uuid)
        response = utils.consume(
            "http://DescargaMasivaTerceros.gob.mx/IAutenticacion/Autentica",
            "https://cfdidescargamasivasolicitud.clouda.sat.gob.mx/Autenticacion/Autenticacion.svc",
            request_content,
        )
        if response.status_code != 200:
            raise exceptions.RequestException(
                response.status_code, response.reason, request_content
            )
        response_clean = utils.remove_namespaces(response.content.decode("UTF-8"))
        self._token = LoginParser.parse(response_clean)["token"]

    def _get_login_soap_body(
        self, created_object: datetime, expires_object: datetime, uuid: str
    ) -> str:
        """Creates the request body to be used in login

        Args:
            created_object (datetime): Creation date to be used in the session
            expires_object (datetime): Expiration date to be used in the session
            uuid (str): UUID date to be used in the session

        Returns:
            str: Content body
        """
        created = created_object.isoformat()
        expires = expires_object.isoformat()
        timestamp = utils.prepare_template(
            TEMPLATES["Timestamp"],
            {
                "created": created,
                "expires": expires,
            },
        )
        digest_value = utils.digest(timestamp)
        signed_info = utils.prepare_template(
            TEMPLATES["SignedInfo"],
            {
                "uri": "#_0",
                "digest_value": digest_value,
            },
        )
        signature_value = self._certificate_handler.sign(signed_info)
        _logger.debug(
            """Creating Login Envelope with the next data
            "created": %s,
            "expires": %s,
            "uuid": %s,
        """,
            created,
            expires,
            uuid,
        )
        envelope = utils.prepare_template(
            TEMPLATES["LoginEnvelope"],
            {
                "binary_security_token": self._certificate_handler.cert,
                "created": created,
                "digest_value": digest_value,
                "expires": expires,
                "signature_value": signature_value,
                "uuid": uuid,
            },
        )
        return envelope



# pylint: disable=redefined-outer-name
import random
from datetime import datetime, timedelta
from importlib import resources

import pytest

import sat.enums
from sat.certificate_handler import CertificateHandler
from sat.sat_connector import SATConnector
from sat.sat_login_handler import SATLoginHandler

from . import fake_fiel, real_fiel

# cert = resources.read_binary(fake_fiel, "EKU9003173C9.cer")
# key = resources.read_binary(fake_fiel, "EKU9003173C9.key")
# password = resources.read_text(fake_fiel, "EKU9003173C9.txt").encode("utf-8")

cert = resources.read_binary(real_fiel, "NAPM9608096N8.cer")
key = resources.read_binary(real_fiel, "NAPM9608096N8.key")
password = resources.read_text(real_fiel, "NAPM9608096N8.txt").encode("utf-8")


@pytest.fixture
def certificate_handler():
    new_certificate_handler = CertificateHandler(cert, key, password)
    return new_certificate_handler


@pytest.fixture
def login_handler(certificate_handler):
    _login_handler = SATLoginHandler(certificate_handler)
    return _login_handler


@pytest.fixture
def sat_connector():
    sat_obj = SATConnector(cert, key, password)
    return sat_obj


query_scenarios = [  # TODO make flexible
    (sat.enums.DownloadType.ISSUED, sat.enums.RequestType.CFDI),
    (sat.enums.DownloadType.RECEIVED, sat.enums.RequestType.CFDI),
    (sat.enums.DownloadType.ISSUED, sat.enums.RequestType.METADATA),
    (sat.enums.DownloadType.RECEIVED, sat.enums.RequestType.METADATA),
]


@pytest.fixture(params=query_scenarios)
def query(sat_connector, request):
    start = datetime.fromisoformat("2021-01-01T00:00:00")
    end = datetime.fromisoformat("2021-03-01T00:00:00") + timedelta(
        seconds=random.randint(1, 10000)
    )
    query = sat_connector.query(
        start=start,
        end=end,
        download_type=request.param[0],
        request_type=request.param[1],
    )
    return query


@pytest.fixture
def package_ids(sat_connector, query):
    package_ids = sat_connector.wait_query(query["IdSolicitud"])
    return package_ids



from datetime import datetime

import pytest

from sat.exceptions import RequestException


class FakeLogin:
    created = datetime.fromisoformat("2019-08-01 03:38:19.000")
    expires = datetime.fromisoformat("2019-08-01 03:43:19.000")
    uuid = "uuid-cf6c80fb-00ae-44c0-af56-54ec65decbaa-1"


@pytest.fixture
def test_login(login_handler):
    login_handler.login()


def test_reuse_token(login_handler):
    login_handler.login()
    first_token = login_handler.token
    login_handler.login()
    second_token = login_handler.token
    assert first_token == second_token


def test_login_expired(login_handler):
    with pytest.raises(RequestException):  # Expired session
        login_handler.login(
            created=FakeLogin.created,
            expires=FakeLogin.expires,
            uuid=FakeLogin.uuid,
        )



# pylint: disable=redefined-outer-name
class FakeSignature:
    data = "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua."
    signature = "afxUybh/NrIf4LTFrOT22wS3mjdurwx6J1gNlYSkNbcmt9NF1jDyw1oT3Akk17bsLUeyH+JQkiii\nydW2Tpi4xmKt1X5JjzVu/zDu/u8R910nn7ZRdVACT95pGuKTa74NoDNW9eX2IruQK8X7TmSHasiE\nSTdzmKZNuLE5VTn2c9O5XQmJUCxSV5oieWHyI02OKJu7D60cT/Ma+yDFPuOSc6pnCY71QH21X3rN\nS8GBHfkgNaC2HA6BpNUGrxTBkcpvTvdyXgPyffwWKQjGyWdRNctnJUy21O6mEJmz6evs93/CAGAa\nLWvv7Q4e+CeaT63hQiVWxjTKY+BN7Uc4b2KgTQ==\n"
    bad_signature = "InvalidSignature"


def test_sign(certificate_handler):
    assert certificate_handler.sign(FakeSignature.data) == FakeSignature.signature


def test_sign_invalid(certificate_handler):
    assert certificate_handler.sign(FakeSignature.data) != FakeSignature.bad_signature



# -*- coding: utf-8 -*-
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="sat-ws",
    version="2.0.0",
    author="Moisés Navarro",  # TODO
    author_email="moisalejandro@gmail.com",  # TODO
    description="API to connect with SAT ws",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/HomebrewSoft/sat_ws_api",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    include_package_data=True,
    python_requires=">=3.6",
    install_requires=[
        "pyOpenSSL==19.1.0",
        "requests==2.24.0",
        "xmltodict==0.12.0",
    ],
)
